@@ -19,12 +19,13 @@ from _pytest._code.code import ExceptionInfo
 from _pytest._code.code import ReprFileLocation
 from _pytest._code.code import TerminalRepr
 from _pytest.compat import safe_getattr
+from _pytest.compat import TYPE_CHECKING
 from _pytest.fixtures import FixtureRequest
 from _pytest.outcomes import Skipped
 from _pytest.python_api import approx
 from _pytest.warning_types import PytestWarning
 
-if False:  # TYPE_CHECKING
+if TYPE_CHECKING:
     import doctest
     from typing import Type
 
@@ -445,7 +446,9 @@ class DoctestModule(pytest.Module):
                     obj = getattr(obj, "fget", obj)
                 return doctest.DocTestFinder._find_lineno(self, obj, source_lines)
 
-            def _find(self, tests, obj, name, module, source_lines, globs, seen):
+            def _find(
+                self, tests, obj, name, module, source_lines, globs, seen
+            ) -> None:
                 if _is_mocked(obj):
                     return
                 with _patch_unwrap_mock_aware():
